# %%
import torch
import yucca
import wandb
import logging
import numpy as np
import os
import matplotlib.pyplot as plt
from batchgenerators.utilities.file_and_folder_operations import join
from torchmetrics import MetricCollection
from torchmetrics.regression import MeanAbsoluteError
from yucca.modules.optimization.loss_functions.deep_supervision import DeepSupervisionLoss
from yucca.functional.utils.files_and_folders import recursive_find_python_class
from yucca.functional.utils.kwargs import filter_kwargs
from yucca.modules.metrics.training_metrics import Accuracy, AUROC, GeneralizedDiceScore
from yucca.functional.visualization import get_train_fig_with_inp_out_tar
from yucca.modules.lightning_modules.BaseLightningModule import BaseLightningModule
from yucca.functional.utils.torch_utils import measure_FLOPs
from fvcore.nn import flop_count_table


class YuccaLightningModule(BaseLightningModule):
    """
    The YuccaLightningModule class is an implementation of the PyTorch Lightning module designed for the Yucca project.
    It extends the LightningModule class and encapsulates the neural network model, loss functions, and optimization logic.
    This class is responsible for handling training, validation, and inference steps within the Yucca machine learning pipeline.
    """

    def __init__(
        self,
        config: dict,
        deep_supervision: bool = False,
        disable_inference_preprocessing: bool = False,
        loss_fn: str = None,
        loss_kwargs: dict = {
            "soft_dice_kwargs": {"apply_softmax": True},
        },
        lr_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.CosineAnnealingLR,
        lr_scheduler_kwargs: dict = {
            "eta_min": 1e-9,
        },
        model_kwargs: dict = {},
        optimizer: torch.optim.Optimizer = torch.optim.SGD,
        optimizer_kwargs={
            "lr": 1e-3,
        },
        sliding_window_overlap: float = 0.5,
        step_logging: bool = False,
        test_time_augmentation: bool = False,
        progress_bar: bool = False,
        log_image_every_n_epochs: int = None,
    ):
        model = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "modules", "networks")],
            class_name=config["model_name"],
            current_module="yucca.modules.networks",
        )
        loss_fn = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "modules", "optimization", "loss_functions")],
            class_name=loss_fn if loss_fn is not None else "DiceCE",
            current_module="yucca.modules.optimization.loss_functions",
        )
        preprocessor = self.get_preprocessor(config)
        super().__init__(
            model=model,
            model_dimensions=config["model_dimensions"],
            num_classes=config["num_classes"],
            num_modalities=config["num_modalities"],
            patch_size=config["patch_size"],
            plans=config["plans"],
            deep_supervision=deep_supervision,
            disable_inference_preprocessing=disable_inference_preprocessing,
            hparams_path=config["plans_path"],
            loss_fn=loss_fn,
            loss_kwargs=loss_kwargs,
            lr_scheduler=lr_scheduler,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            model_kwargs=model_kwargs,
            optimizer=optimizer,
            optimizer_kwargs=optimizer_kwargs,
            preprocessor=preprocessor,
            progress_bar=progress_bar,
            sliding_window_overlap=sliding_window_overlap,
            sliding_window_prediction=config["patch_based_training"],
            step_logging=step_logging,
            test_time_augmentation=test_time_augmentation,
        )
        self.config = config
        self.task_type = config["task_type"]
        self.log_image_every_n_epochs = log_image_every_n_epochs

        self.use_label_regions = "use_label_regions" in config.keys() and config["use_label_regions"]
        # If we are training we save params and then start training
        # Do not overwrite parameters during inference.
        self.save_hyperparameters(ignore=["lr_scheduler", "optimizer"])

    def setup(self, stage):  # noqa: U100
        logging.info(f"Loading Model: {self.model_dimensions} {self.model.__name__}")
        if self.model_dimensions == "3D":
            conv_op = torch.nn.Conv3d
            norm_op = torch.nn.InstanceNorm3d
        else:
            conv_op = torch.nn.Conv2d
            norm_op = torch.nn.BatchNorm2d

        model_kwargs = {
            # Applies to most CNN-based architectures
            "conv_op": conv_op,
            "deep_supervision": self.deep_supervision,
            "norm_op": norm_op,
            # UNetR
            "patch_size": self.patch_size,
            # MedNeXt
            "checkpoint_style": "outside_block",
        }
        model_kwargs.update(self.model_kwargs)
        model_kwargs = filter_kwargs(self.model, model_kwargs)
        self.model = self.model(input_channels=self.num_modalities, num_classes=self.num_classes, **model_kwargs)
        self.visualize_model_with_FLOPs()

    def visualize_model_with_FLOPs(self):
        try:
            data = torch.randn((self.config["batch_size"], self.config["num_modalities"], *self.config["patch_size"]))
            flops = measure_FLOPs(self.model, data)
            del data
            logging.info("\n" + flop_count_table(flops))
        except RuntimeError:
            logging.info("\n Model architecture could not be visualized.")

    def on_fit_start(self):
        if self.task_type == "classification":
            tmetrics_task = "multiclass" if self.num_classes > 2 else "binary"
            # can we get per-class?
            self.train_metrics = MetricCollection(
                {
                    "train/acc": Accuracy(task=tmetrics_task, num_classes=self.num_classes),
                    "train/roc_auc": AUROC(task=tmetrics_task, num_classes=self.num_classes),
                }
            )
            self.val_metrics = MetricCollection(
                {
                    "val/acc": Accuracy(task=tmetrics_task, num_classes=self.num_classes),
                    "val/roc_auc": AUROC(task=tmetrics_task, num_classes=self.num_classes),
                }
            )

        if self.task_type == "segmentation":
            self.train_metrics = MetricCollection(
                {
                    "train/aggregated_dice": GeneralizedDiceScore(
                        multilabel=self.use_label_regions,
                        num_classes=self.num_classes,
                        include_background=self.num_classes == 1,
                        weight_type="linear",
                        per_class=False,
                    ),
                    "train/dice": GeneralizedDiceScore(
                        multilabel=self.use_label_regions,
                        num_classes=self.num_classes,
                        include_background=self.num_classes == 1,
                        weight_type="linear",
                        per_class=True,
                    ),
                },
            )

            self.val_metrics = MetricCollection(
                {
                    "val/aggregated_dice": GeneralizedDiceScore(
                        multilabel=self.use_label_regions,
                        num_classes=self.num_classes,
                        include_background=self.num_classes == 1,
                        weight_type="linear",
                        per_class=False,
                    ),
                    "val/dice": GeneralizedDiceScore(
                        multilabel=self.use_label_regions,
                        num_classes=self.num_classes,
                        include_background=self.num_classes == 1,
                        weight_type="linear",
                        per_class=True,
                    ),
                },
            )

        if self.task_type == "self-supervised":
            self.train_metrics = MetricCollection({"train/MAE": MeanAbsoluteError()})
            self.val_metrics = MetricCollection({"train/MAE": MeanAbsoluteError()})

        if self.log_image_every_n_epochs is None:
            self.log_image_every_n_epochs = self.get_image_logging_epochs(self.trainer.max_epochs)

    def training_step(self, batch, batch_idx):
        inputs, target, file_path = batch["image"], batch["label"], batch["file_path"]
        output = self(inputs)
        loss = self.loss_fn_train(output, target)

        if self.deep_supervision:
            # If deep_supervision is enabled output and target will be a list of (downsampled) tensors.
            # We only need the original ground truth and its corresponding prediction which is always the first entry in each list.
            output = output[0]
            target = target[0]

        metrics = self.compute_metrics(self.train_metrics, output, target)
        self.log_dict(
            {"train/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )

        if batch_idx == 0 and wandb.run is not None and self.log_image_this_epoch is True:
            self._log_dict_of_images_to_wandb(
                {
                    "input": inputs.detach().cpu().to(torch.float32).numpy(),
                    "target": target.detach().cpu().to(torch.float32).numpy(),
                    "output": output.detach().cpu().to(torch.float32).numpy(),
                    "file_path": file_path,
                },
                log_key="train",
                task_type=self.task_type,
            )

        return loss

    def validation_step(self, batch, batch_idx):
        inputs, target, file_path = batch["image"], batch["label"], batch["file_path"]
        output = self(inputs)
        loss = self.loss_fn_val(output, target)

        metrics = self.compute_metrics(self.val_metrics, output, target)
        self.log_dict(
            {"val/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )

        if batch_idx == 0 and wandb.run is not None and self.log_image_this_epoch is True:
            self._log_dict_of_images_to_wandb(
                {
                    "input": inputs.detach().cpu().to(torch.float32).numpy(),
                    "target": target.detach().cpu().to(torch.float32).numpy(),
                    "output": output.detach().cpu().to(torch.float32).numpy(),
                    "file_path": file_path,
                },
                log_key="val",
                task_type=self.task_type,
            )

    def configure_optimizers(self):
        # Configure the loss function
        loss_kwargs = filter_kwargs(self.loss_fn, self.loss_kwargs)
        self.loss_fn_train = self.loss_fn(**loss_kwargs)
        self.loss_fn_val = self.loss_fn(**loss_kwargs)
        if self.deep_supervision:
            self.loss_fn_train = DeepSupervisionLoss(self.loss_fn_train, weights=None)

        # Configure the optimizer
        optim_kwargs = {
            "momentum": 0.9,
            "eps": 1e-4,
            "weight_decay": 3e-5,
        }
        optim_kwargs.update(self.optim_kwargs)
        optim_kwargs = filter_kwargs(self.optim, optim_kwargs)
        self.optim = self.optim(self.model.parameters(), **optim_kwargs)

        # Configure the lr scheduler
        lr_scheduler_kwargs = {
            "T_max": self.trainer.max_epochs,
        }
        lr_scheduler_kwargs.update(self.lr_scheduler_kwargs)
        lr_scheduler_kwargs = filter_kwargs(self.lr_scheduler, lr_scheduler_kwargs)
        self.lr_scheduler = self.lr_scheduler(self.optim, **lr_scheduler_kwargs)

        # Finally return the optimizer and scheduler - the loss is not returned.
        return {"optimizer": self.optim, "lr_scheduler": self.lr_scheduler}

    def _log_dict_of_images_to_wandb(self, imagedict: {}, log_key: str, task_type: str = "segmentation"):
        batch_idx = np.random.randint(0, imagedict["input"].shape[0])
        case = os.path.splitext(os.path.split(imagedict["file_path"][batch_idx])[-1])[0]

        fig = get_train_fig_with_inp_out_tar(
            input=imagedict["input"][batch_idx],
            output=imagedict["output"][batch_idx],
            target=imagedict["target"][batch_idx],
            fig_title=case,
            task_type=task_type,
        )
        wandb.log({log_key: wandb.Image(fig)}, commit=False)
        plt.close(fig)

    @staticmethod
    def get_preprocessor(config):
        preprocessor = None
        if config.get("plans") and config["plans"].get("preprocessor"):
            preprocessor = recursive_find_python_class(
                folder=[join(yucca.__path__[0], "pipeline", "preprocessing")],
                class_name=config["plans"]["preprocessor"],
                current_module="yucca.pipeline.preprocessing",
            )
        return preprocessor

    @property
    def log_image_this_epoch(self):
        if isinstance(self.log_image_every_n_epochs, int):
            return self.current_epoch % self.log_image_every_n_epochs == 0
        if isinstance(self.log_image_every_n_epochs, list):
            return self.current_epoch in self.log_image_every_n_epochs

    @staticmethod
    def get_image_logging_epochs(final_epoch: int = 1000):
        first_half = np.logspace(0, 5, 10, base=4, endpoint=False)
        second_half = final_epoch - np.logspace(0, 5, 10, base=4, endpoint=False)[::-1]
        indices = sorted(np.concatenate((first_half, second_half)).astype(int))
        return indices


if __name__ == "__main__":
    f = YuccaLightningModule(
        config={
            "model_name": "TinyUNet",
            "plans": {"preprocessor": "YuccaPreprocessor"},
            "model_dimensions": "3D",
            "num_classes": 2,
            "num_modalities": 1,
            "patch_size": (32, 32, 32),
            "plans_path": "",
            "patch_based_training": True,
            "task_type": "segmentation",
        },
    )
    data = torch.randn((2, 1, *(32, 32, 32)))
    f.setup(stage="test")
    f.forward(data)

# %%



from batchgenerators.utilities.file_and_folder_operations import subfiles, join, maybe_mkdir_p as ensure_dir_exists
from SimpleITK import ReadImage
import glob
import os
import logging
import SimpleITK as sitk


"""
Task031_MSD_BrainTumour
Task032_MSD_Heart
Task033_MSD_Liver
Task034_MSD_Hippocampus
Task035_MSD_Prostate
Task036_MSD_Lung
Task037_MSD_Pancreas
Task038_MSD_HepaticVessel
Task039_MSD_Spleen
Task040_MSD_Colon
"""


current = [
    "Task031_MSD_BrainTumour",
    "Task032_MSD_Heart",
    "Task033_MSD_Liver",
    "Task034_MSD_Hippocampus",
    "Task035_MSD_Prostate",
    "Task036_MSD_Lung",
    "Task037_MSD_Pancreas",
    "Task038_MSD_HepaticVessel",
    "Task039_MSD_Spleen",
    "Task040_MSD_Colon",
]

expected = [
    "Task01_BrainTumour",
    "Task02_Heart",
    "Task03_Liver",
    "Task04_Hippocampus",
    "Task05_Prostate",
    "Task06_Lung",
    "Task07_Pancreas",
    "Task08_HepaticVessel",
    "Task09_Spleen",
    "Task10_Colon",
]

folders = [
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task031_MSD_BrainTumour/Task031_MSD_BrainTumour/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_4/version_0/best",
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task032_MSD_Heart/Task032_MSD_Heart/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_4/version_0/best",
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task033_MSD_Liver/Task033_MSD_Liver/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_4/version_0/best",
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task034_MSD_Hippocampus/Task034_MSD_Hippocampus/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_4/version_0/best",
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task035_MSD_Prostate/Task035_MSD_Prostate/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_3/version_0/best",
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task036_MSD_Lung/Task036_MSD_Lung/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_4/version_0/best",
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task037_MSD_Pancreas/Task037_MSD_Pancreas/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_4/version_0/best",
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task038_MSD_HepaticVessel/Task038_MSD_HepaticVessel/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_4/version_0/best",
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task039_MSD_Spleen/Task039_MSD_Spleen/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_4/version_0/best",
    "/projects/zcr545_shared_datasets/people/zcr545/YuccaData/yucca_results/Task040_MSD_Colon/Task040_MSD_Colon/UNet__3D/YuccaManagerV11__YuccaPlannerV3/default/kfold_5_fold_4/version_0/best",
]

exclude = [
    "Task03_Liver/liver_141.nii.gz",
    "Task03_Liver/liver_156.nii.gz",
    "Task03_Liver/liver_160.nii.gz",
    "Task03_Liver/liver_161.nii.gz",
    "Task03_Liver/liver_162.nii.gz",
    "Task03_Liver/liver_164.nii.gz",
    "Task03_Liver/liver_167.nii.gz",
    "Task03_Liver/liver_182.nii.gz",
    "Task03_Liver/liver_189.nii.gz",
    "Task03_Liver/liver_190.nii.gz",
    "Task08_HepaticVessel/hepaticvessel_247.nii.gz",
]

savepath = "/home/zcr545/decathlon_submission"
ensure_dir_exists(savepath)

blacklist = [
    "Task03_Liver/liver_141.nii.gz",
    "Task03_Liver/liver_156.nii.gz",
    "Task03_Liver/liver_160.nii.gz",
    "Task03_Liver/liver_161.nii.gz",
    "Task03_Liver/liver_162.nii.gz",
    "Task03_Liver/liver_164.nii.gz",
    "Task03_Liver/liver_167.nii.gz",
    "Task03_Liver/liver_182.nii.gz",
    "Task03_Liver/liver_189.nii.gz",
    "Task03_Liver/liver_190.nii.gz",
    "Task08_HepaticVessel/hepaticvessel_247.nii.gz",
]

for i in range(10):
    for seg_path in subfiles(folders[i], join=False):
        new_filename = seg_path[len(current[i]) :]
        new_path = join(expected[i], new_filename)
        if new_path in blacklist:
            print("blocking: ", new_path)
            continue
        ensure_dir_exists(join(savepath, expected[i]))


def copy_geometry(image: sitk.Image, ref: sitk.Image):
    if ref.GetDimension() == 4:
        ref = ref[:, :, :, 0]
    print(image.GetSize(), ref.GetSize(), ref.GetDimension())
    image.SetOrigin(ref.GetOrigin())
    image.SetDirection(ref.GetDirection())
    image.SetSpacing(ref.GetSpacing())
    return image


logger = logging.getLogger()
sub_dir = "/home/zcr545/decathlon_submission"
gt_dir = "/home/zcr545/data/data/public_datasets/decathlon"


for i in range(10):
    img_list = [x for x in glob.glob(os.path.join(sub_dir, expected[i], "*.nii.gz"))]
    for predpath in img_list:
        pred = ReadImage(predpath)
        name = os.path.split(predpath)[-1]
        gtpath = os.path.join(gt_dir, expected[i], "imagesTs", name)
        gt = ReadImage(gtpath)
        copy_geometry(pred, gt)
# %%



import numpy as np
import os
import shutil
import nibabel as nib
from yucca.paths import get_models_path, get_preprocessed_data_path, get_raw_data_path
from typing import Literal
from batchgenerators.utilities.file_and_folder_operations import save_json, subfiles, join, subdirs
from tqdm import tqdm
from pathlib import Path


def combine_images_from_tasks(tasks: list, target_base: str, run_type: Literal["supervised", "unsupervised"]):
    assert len(tasks) > 0, "list of tasks empty"
    for task in tqdm(tasks):
        folders = ["imagesTr", "imagesTs", "labelsTr", "labelsTs"] if run_type == "supervised" else ["imagesTr"]
        for folder in folders:
            source = os.path.join(get_raw_data_path(), task, folder)
            target = os.path.join(target_base, folder)
            print("Copying ", source, target)
            copy_files_from_to(source, target)


def copy_files_from_to(source_dir, target_dir):
    assert os.path.isdir(source_dir)
    os.makedirs(target_dir, exist_ok=True)

    for file in tqdm(os.listdir(source_dir)):
        shutil.copy2(os.path.join(source_dir, file), f"{target_dir}/{file}")


def get_identifiers_from_splitted_files(folder: str, ext, tasks: list):
    if len(tasks) > 0:
        uniques = np.unique(
            [i[: -len("_000." + ext)] for task in tasks for i in subfiles(join(folder, task, suffix=ext), join=False)]
        )
    else:
        uniques = np.unique([i[: -len("_000." + ext)] for i in subfiles(folder, suffix=ext, join=False)])
    return list(uniques)


def dirs_in_dir(dir: str):
    p = Path(dir)
    return [f.name for f in p.iterdir() if f.is_dir() and f.name[0] not in [".", "_"]]


def files_in_dir(dir: str):
    p = Path(dir)
    return [f.name for f in p.iterdir() if f.is_file() and f.name[0] not in [".", "_"]]


def should_use_volume(vol: nib.Nifti1Image):
    return not (np.any(np.array(vol.shape) < 15) or len(vol.shape) != 3 or np.array(vol.dataobj).min() < 0)


def remove_punctuation_and_spaces(data: str):
    data = data.replace(" ", "-").replace(",", "-").replace(".", "-")
    return data


def generate_dataset_json(
    output_file: str,
    imagesTr_dir: str,
    imagesTs_dir: str,
    modalities: dict,
    labels: dict,
    dataset_name: str,
    label_hierarchy: dict = {},
    regions: dict = {},
    tasks: list = [],
    license: str = "hands off!",
    dataset_description: str = "",
    dataset_reference="",
    dataset_release="0.0",
):
    """
    :param output_file: This needs to be the full path to the dataset.json you intend to write, so
    output_file='DATASET_PATH/dataset.json' where the folder DATASET_PATH points to is the one with the
    imagesTr and labelsTr subfolders
    :param imagesTr_dir: path to the imagesTr folder of that dataset
    :param imagesTs_dir: path to the imagesTs folder of that dataset. Can be None
    :param modalities: dict of modality names and their corresponding values. must be in the same order as the images (first entry
    corresponds to _000.nii.gz, etc). Example: ('T1', 'T2', 'FLAIR').
    :param labels: dict with int->str (key->value) mapping the label IDs to label names. Note that 0 is always
    supposed to be background! Example: {0: 'background', 1: 'left hippocampus', 2: 'right hippocampus'}
    :param regions: dict with str -> dict where inner dict contains specification of each region
                    The dict must have the following form:
                    ```
                    {
                        'REGION_NAME': {
                            'priority': 1, # LOWER IS MORE IMPORTANT
                            'labels': ['LABEL1', 'LABEL2']
                        }
                        ...
                    }
                    where
                        priority specifies how regions are collapsed during prediction
                        labels specify which labels to be included into a region.
                               each label must be a key in the labels dict above.
    :param dataset_name: The name of the dataset. Can be anything you want
    :param license:
    :param dataset_description:
    :param dataset_reference: website of the dataset, if available
    :param dataset_release:
    :return:
    """
    first_file = files_in_dir(imagesTr_dir)[0]
    im_ext = os.path.split(first_file)[-1].split(os.extsep, 1)[-1]
    train_identifiers = get_identifiers_from_splitted_files(imagesTr_dir, im_ext, tasks)

    if imagesTs_dir is not None:
        test_identifiers = get_identifiers_from_splitted_files(imagesTs_dir, im_ext, tasks)
    else:
        test_identifiers = []

    labels = {str(i): labels[i] for i in labels.keys()} if labels is not None else None

    # Sanity check regions
    if regions is not None:
        assert labels is not None
        assert isinstance(regions, dict), "Regions must be specified using a dict mapping regions to labels"
        priorities = []
        for region, region_dict in regions.items():
            assert "priority" in region_dict.keys(), f"Region {region} must have a priority specifier"
            priorities.append(region_dict["priority"])
            assert "labels" in region_dict.keys(), f"Region {region} must have labels"
            assert isinstance(region_dict["labels"], list), f"Region {region} labels must be a list"
            for label in region_dict["labels"]:
                assert isinstance(label, str), f"label must be of type str not {type(label)}"
                assert (
                    label in labels.values()
                ), f"Labels must be strings, corresponding to the values in the label dict: {labels.values()}"
                assert len(region_dict.keys()) == 2, region_dict.keys()

        # check that the priorities in the dict contains all numbers between 1 and max(priorities):
        assert set(priorities) == set(
            range(1, len(regions) + 1)
        ), f"The regions dict must contain all priorities between 1 and len(regions), but instead got priorities: {priorities}"

    json_dict = {}
    json_dict["name"] = dataset_name
    json_dict["description"] = dataset_description
    json_dict["tensorImageSize"] = "4D"
    json_dict["reference"] = dataset_reference
    json_dict["licence"] = license
    json_dict["release"] = dataset_release
    json_dict["image_extension"] = im_ext
    json_dict["modality"] = {str(i): modalities[i] for i in range(len(modalities))}
    json_dict["labels"] = labels
    json_dict["label_hierarchy"] = label_hierarchy
    json_dict["regions"] = regions
    json_dict["tasks"] = tasks
    json_dict["numTraining"] = len(train_identifiers)
    json_dict["numTest"] = len(test_identifiers)
    json_dict["training"] = [{"image": name, "label": name if labels else None} for name in train_identifiers]
    json_dict["test"] = test_identifiers

    if not output_file.endswith("dataset.json"):
        print(
            "WARNING: output file name is not dataset.json! This may be intentional or not. You decide. "
            "Proceeding anyways..."
        )
    save_json(json_dict, os.path.join(output_file))


def get_task_from_task_id(task_id: str | int, stage: str):
    assert stage in ["raw", "preprocessed", "models"], stage
    task_id = str(task_id)

    if stage == "raw":
        stage_path = get_raw_data_path()
    elif stage == "preprocessed":
        stage_path = get_preprocessed_data_path()
    elif stage == "models":
        stage_path = get_models_path()

    tasks = subdirs(stage_path, join=False)

    # Check if name is already complete
    if task_id in tasks:
        return task_id

    # If not, we try to recreate the name
    # We use the raw_data folder as reference
    for task in tasks:
        if task_id.lower() in task.lower():
            return task

    raise LookupError(f"Task {task_id} not found in {stage_path}.")



import nibabel as nib
import shutil
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p as ensure_dir_exists, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import get_raw_data_path, get_source_path
from yucca.functional.testing.data.nifti import verify_spacing_is_equal, verify_orientation_is_equal


def convert(path: str = get_source_path(), subdir: str = "decathlon", subsubdir: str = "Task06_Lung"):
    # INPUT DATA
    path = f"{path}/{subdir}/{subsubdir}"
    file_suffix = ".nii.gz"

    # OUTPUT DATA
    # Define the task name and prefix
    task_name = "Task036_MSD_Lung"

    # Set target paths
    target_base = join(get_raw_data_path(), task_name)
    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")
    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    ensure_dir_exists(target_imagesTr)
    ensure_dir_exists(target_labelsTs)
    ensure_dir_exists(target_imagesTs)
    ensure_dir_exists(target_labelsTr)

    # Split data
    images_dir_tr = join(path, "imagesTr")
    labels_dir_tr = join(path, "labelsTr")
    images_dir_ts = join(path, "imagesTs")

    # Populate Target Directory
    # This is also the place to apply any re-orientation, resampling and/or label correction.

    for sTr in subfiles(images_dir_tr, join=False):
        image_path = join(images_dir_tr, sTr)
        label_path = join(labels_dir_tr, sTr)
        sTr = sTr[: -len(file_suffix)]

        image = nib.load(image_path)
        label = nib.load(label_path)
        assert verify_spacing_is_equal(image, label), "spacing"
        assert verify_orientation_is_equal(image, label), "orientation"

        shutil.copy2(image_path, f"{target_imagesTr}/{sTr}_000.nii.gz")
        shutil.copy2(label_path, f"{target_labelsTr}/{sTr}.nii.gz")

    for sTs in subfiles(images_dir_ts, join=False):
        image_path = join(images_dir_ts, sTs)
        sTs = sTs[: -len(file_suffix)]

        image = nib.load(image_path)

        shutil.copy2(image_path, f"{target_imagesTs}/{sTs}_000.nii.gz")

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("CT",),
        labels={0: "Background", 1: "Cancer"},
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="Decathlon: Lung and cancer segmentation",
        dataset_reference="The Cancer Imaging Archive",
    )



import numpy as np


def translate_region_labels(regions: dict, labels: dict) -> dict:
    str_label_to_int = {v: int(k) for k, v in labels.items()}
    regions_with_int_labels = {}
    for region, region_dict in regions.items():
        regions_with_int_labels[region] = {
            "priority": region_dict["priority"],
            "labels": [str_label_to_int[label] for label in region_dict["labels"]],
        }
    return regions_with_int_labels


def batch_convert_labels_to_regions(label, regions):
    b, c, *shape = label.shape
    assert c == 1, f"Class labels are not onehot encoded. Channel dim must be 1, but got c={c}."
    region_canvas = np.zeros((b, len(regions), *shape))

    for channel, region in enumerate(regions.keys()):
        region_labels = regions[region]["labels"]
        assert isinstance(region_labels[0], int)
        region_canvas[:, channel] = np.isin(label[:, 0], region_labels)
    region_canvas = region_canvas.astype(np.uint8)

    return region_canvas


def convert_labels_to_regions(data, regions):
    c, *shape = data.shape
    assert (
        c == 1
    ), f"# Channels is not 1. Make sure the input to this function is a segmentation map of dims (1,h,w[,d]), found shape: {data.shape}"
    region_canvas = np.zeros((len(regions), *shape))
    for channel, region in enumerate(regions):
        region_labels = regions[region]["labels"]
        assert isinstance(region_labels[0], int)
        region_canvas[channel] = np.isin(data[0], region_labels)
    region_canvas = region_canvas.astype(np.uint8)
    return region_canvas


def batch_convert_regions_to_labels(data, region_labels):
    print(
        "You are using the batch_convert_regions_to_labels function."
        "Currently it does not support the priority key. Use with caution."
    )
    b, _, *shape = data.shape
    region_canvas = np.zeros((b, 1, *shape))
    for channel, region_label in enumerate(region_labels):
        mask = data[:, channel] > 0.5
        region_canvas[:, 0][mask] = region_label
    region_canvas = region_canvas
    return region_canvas


def convert_regions_to_labels(data, region_labels):
    print(
        "You are using the convert_regions_to_labels function."
        "Currently it does not support the priority key. Use with caution."
    )
    _, *shape = data.shape
    region_canvas = np.zeros((1, *shape))
    for channel, region_label in enumerate(region_labels):
        mask = data[channel] > 0.5
        region_canvas[0][mask] = region_label
    region_canvas = region_canvas
    return region_canvas



import sys
import numpy as np
import nibabel as nib
import logging
from typing import Optional
from yucca.functional.transforms.label_transforms import convert_labels_to_regions, translate_region_labels
from yucca.functional.utils.nib_utils import get_nib_spacing
from yucca.functional.evaluation.obj_metrics import get_obj_stats_for_label
from yucca.functional.evaluation.surface_metrics import get_surface_metrics_for_label
from tqdm import tqdm
from batchgenerators.utilities.file_and_folder_operations import join
from sklearn.metrics import confusion_matrix
from yucca.functional.evaluation.metrics import auroc


def evaluate_multilabel_folder_segm(
    labels: dict,
    metrics,
    subjects,
    folder_with_predictions,
    folder_with_ground_truth,
    as_binary: Optional[bool] = False,
    obj_metrics: Optional[bool] = False,
    regions: Optional[list] = None,
    surface_metrics: Optional[bool] = False,
    surface_tol=1,
):
    # predictions are multilabel at this point (c, h, w, d)
    # ground truth MAY be converted, but may also be (h, w, d) in which case we use the label_regions
    # to convert for multilabel evaluation
    logging.info(f"Multilabel segmentation evaluation with regions: {regions} and labels: {labels}")

    sys.stdout.flush()
    resultdict = {}
    meandict = {}

    labelarr = np.array(range(len(regions.keys()) + 1), dtype=np.uint8)

    for label in regions.keys():
        meandict[str(label)] = {k: [] for k in list(metrics.keys()) + obj_metrics + surface_metrics}

    for case in tqdm(subjects, desc="Evaluating"):
        casedict = {}
        predpath = join(folder_with_predictions, case)
        gtpath = join(folder_with_ground_truth, case)

        pred = nib.load(predpath)
        spacing = get_nib_spacing(pred)[:3]
        pred = pred.get_fdata().astype(np.uint8)
        pred = pred.transpose([3, 0, 1, 2])
        gt = nib.load(gtpath).get_fdata()

        if len(pred.shape) == len(gt.shape) + 1:
            # In thise case gt has not been converted to multilabel
            assert (
                regions is not None
            ), "Regions must be supplied if ground truth is not already multilabel (i.e. multiple channels)"
            translated_regions = translate_region_labels(regions=regions, labels=labels)
            gt = convert_labels_to_regions(gt[np.newaxis], translated_regions)
            for i in range(len(regions.keys())):
                pred[i] *= 1 + i
                gt[i] *= 1 + i

        if as_binary:
            cmat = confusion_matrix(
                np.around(gt.flatten()).astype(bool).astype(np.uint8),
                np.around(pred.flatten()).astype(bool).astype(np.uint8),
                labels=labelarr,
            )
        else:
            cmat = confusion_matrix(
                np.around(gt.flatten()).astype(np.uint8),
                np.around(pred.flatten()).astype(np.uint8),
                labels=labelarr,
            )

        for label, region_name in enumerate(regions.keys()):
            label += 1
            labeldict = {}

            tp = cmat[label, label]
            fp = sum(cmat[:, label]) - tp
            fn = sum(cmat[label, :]) - tp
            tn = np.sum(cmat) - tp - fp - fn  # often a redundant and meaningless metric
            for k, v in metrics.items():
                labeldict[k] = round(v(tp, fp, tn, fn), 4)
                meandict[str(region_name)][k].append(labeldict[k])

            if obj_metrics:
                raise NotImplementedError
                # now for the object metrics
                # obj_labeldict = get_obj_stats_for_label(gt, pred, label, spacing=spacing, as_binary=as_binary)
                # for k, v in obj_labeldict.items():
                #    labeldict[k] = round(v, 4)
                #    meandict[str(label)][k].append(labeldict[k])

            if surface_metrics:
                if label == 0:
                    surface_labeldict = get_surface_metrics_for_label(
                        gt[label], pred[label], 0, spacing=spacing, tol=surface_tol, as_binary=as_binary
                    )
                else:
                    surface_labeldict = get_surface_metrics_for_label(
                        gt[label - 1], pred[label - 1], label, spacing=spacing, tol=surface_tol, as_binary=as_binary
                    )
                for k, v in surface_labeldict.items():
                    labeldict[k] = round(v, 4)
                    meandict[str(region_name)][k].append(labeldict[k])
            casedict[str(region_name)] = labeldict
        casedict["Prediction:"] = predpath
        casedict["Ground Truth:"] = gtpath

        resultdict[case] = casedict
        del pred, gt, cmat

    for label in regions.keys():
        meandict[str(label)] = {
            k: round(np.nanmean(v), 4) if not np.all(np.isnan(v)) else 0 for k, v in meandict[str(label)].items()
        }
    resultdict["mean"] = meandict
    return resultdict


def evaluate_folder_segm(
    labels,
    metrics,
    subjects,
    folder_with_predictions,
    folder_with_ground_truth,
    as_binary: Optional[bool] = False,
    obj_metrics: Optional[bool] = False,
    surface_metrics: Optional[bool] = False,
    surface_tol=1,
):
    logging.info(f"segmentation evaluation with labels: {labels}")

    sys.stdout.flush()
    resultdict = {}
    meandict = {}

    for label in labels:
        meandict[str(label)] = {k: [] for k in list(metrics.keys()) + obj_metrics + surface_metrics}

    for case in tqdm(subjects, desc="Evaluating"):
        casedict = {}
        predpath = join(folder_with_predictions, case)
        gtpath = join(folder_with_ground_truth, case)

        pred = nib.load(predpath)
        spacing = get_nib_spacing(pred)
        pred = pred.get_fdata()
        gt = nib.load(gtpath).get_fdata()

        if as_binary:
            cmat = confusion_matrix(
                np.around(gt.flatten()).astype(bool).astype(np.uint8),
                np.around(pred.flatten()).astype(bool).astype(np.uint8),
                labels=labels,
            )
        else:
            cmat = confusion_matrix(
                np.around(gt.flatten()).astype(np.uint8),
                np.around(pred.flatten()).astype(np.uint8),
                labels=labels,
            )

        for label in labels:
            labeldict = {}

            tp = cmat[label, label]
            fp = sum(cmat[:, label]) - tp
            fn = sum(cmat[label, :]) - tp
            tn = np.sum(cmat) - tp - fp - fn  # often a redundant and meaningless metric
            for k, v in metrics.items():
                labeldict[k] = round(v(tp, fp, tn, fn), 4)
                meandict[str(label)][k].append(labeldict[k])

            if obj_metrics:
                # now for the object metrics
                obj_labeldict = get_obj_stats_for_label(gt, pred, label, spacing=spacing, as_binary=as_binary)
                for k, v in obj_labeldict.items():
                    labeldict[k] = round(v, 4)
                    meandict[str(label)][k].append(labeldict[k])

            if surface_metrics:
                surface_labeldict = get_surface_metrics_for_label(
                    gt, pred, label, spacing=spacing, tol=surface_tol, as_binary=as_binary
                )
                for k, v in surface_labeldict.items():
                    labeldict[k] = round(v, 4)
                    meandict[str(label)][k].append(labeldict[k])
            casedict[str(label)] = labeldict
        casedict["Prediction:"] = predpath
        casedict["Ground Truth:"] = gtpath

        resultdict[case] = casedict
        del pred, gt, cmat

    for label in labels:
        meandict[str(label)] = {
            k: round(np.nanmean(v), 4) if not np.all(np.isnan(v)) else 0 for k, v in meandict[str(label)].items()
        }
    resultdict["mean"] = meandict
    return resultdict


def evaluate_folder_cls(
    labels,
    metrics,
    subjects,
    folder_with_predictions,
    folder_with_ground_truth,
):
    """
    Evaluate classification results
    """
    sys.stdout.flush()
    resultdict = {}

    predictions = []
    prediction_probs = []
    ground_truths = []

    # Flag to check if we have prediction probabilities to calculate AUROC
    use_probs = False

    # load predictions and ground truths
    for case in tqdm(subjects, desc="Evaluating"):
        predpath = join(folder_with_predictions, case)
        gtpath = join(folder_with_ground_truth, case)

        pred: int = np.loadtxt(predpath)
        gt: int = np.loadtxt(gtpath)

        try:
            if len(prediction_probs) == 0:
                print("Prediction probabilities found. Will use them for evaluation.")
                use_probs = True

            pred_probs = np.load(predpath.replace(".txt", ".npz"))["data"]  # contains output probabilities
            prediction_probs.append(pred_probs)
        except FileNotFoundError:
            pred_probs = None

        predictions.append(pred)
        ground_truths.append(gt)

    predictions = np.array(predictions)
    ground_truths = np.array(ground_truths)

    # calculate per-class metrics
    cmat = confusion_matrix(ground_truths, predictions, labels=labels)

    resultdict["per_class"] = {}

    for label in labels:
        tp = cmat[label, label]
        fp = sum(cmat[:, label]) - tp
        fn = sum(cmat[label, :]) - tp
        tn = np.sum(cmat) - tp - fp - fn

        labeldict = {}

        for k, v in metrics.items():
            labeldict[k] = round(v(tp, fp, tn, fn), 4)

        resultdict["per_class"][str(label)] = labeldict

    # calculate AUROC
    if use_probs:
        auroc_per_class: list[float] = auroc(ground_truths, prediction_probs)
        for label, score in zip(labels, auroc_per_class):
            resultdict["per_class"][str(label)]["AUROC"] = round(score, 4)

    # caclulate global (mean) metrics
    resultdict["mean"] = {}
    for k, _ in resultdict["per_class"][str(labels[0])].items():
        resultdict["mean"][k] = sum([resultdict["per_class"][str(label)][k] for label in labels])
        resultdict["mean"][k] = round(resultdict["mean"][k] / len(labels), 4)

    return resultdict



import numpy as np
from typing import List
from yucca.functional.transforms.label_transforms import batch_convert_labels_to_regions, translate_region_labels
from yucca.modules.data.augmentation.transforms.YuccaTransform import YuccaTransform


class ConvertLabelsToRegions(YuccaTransform):
    def __init__(
        self,
        convert_labels_to_regions=False,
        label_key="label",
        labels: dict[str, str] = None,
        regions: dict[str, dict] = None,
    ):
        self.convert_labels_to_regions = convert_labels_to_regions
        self.label_key = label_key

        if self.convert_labels_to_regions:
            assert regions is not None, "you cannot enable convert_labels_to_regions while not supplying any regions"
            assert labels is not None, "you cannot enable convert_labels_to_regions while not supplying labels"
            # regions stores region labels in str format for readability,
            # but here we need the actual label integer from the label map
            # so we translate it to integers
            self.regions = translate_region_labels(regions, labels)

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __convert__(self, label: np.ndarray, regions: List[List[int]]):
        return batch_convert_labels_to_regions(label, regions)

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        if self.convert_labels_to_regions:
            assert (
                len(data_dict[self.label_key].shape) == 5 or len(data_dict[self.label_key].shape) == 4
            ), f"Incorrect data size or shape.\
                \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.label_key].shape}"
            data_dict[self.label_key] = self.__convert__(data_dict[self.label_key], self.regions)
        return data_dict



config = {
    "batch_size": 2,
    "crop_to_nonzero": True,
    "continue_from_most_recent": True,
    "dims": "2D",
    "deep_supervision": False,
    "experiment": "default",
    "extension": ".nii.gz",
    "learning_rate": 1e-3,
    "loss_fn": "DiceCE",
    "max_epochs": 2,
    "modalities": ("MRI",),
    "model_name": "TinyUNet",
    "momentum": 0.99,
    "norm_op": "volume_wise_znorm",
    "num_classes": 3,
    "num_modalities": 1,
    "patch_size": (32, 32),
    "plans_name": "demo",
    "plans": None,
    "split_idx": 0,
    "split_method": "kfold",
    "split_param": 5,
    "task": "Task000_TEST_SEGMENTATION",
    "task_type": "segmentation",
}



import torch
from yucca.pipeline.managers.YuccaManager import YuccaManager
from yucca.modules.data.augmentation.augmentation_presets import genericV2


class YuccaManagerV2(YuccaManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.deep_supervision = True
        self.augmentation_params = genericV2
        self.optimizer = torch.optim.AdamW
        self.optim_kwargs = {"eps": 1e-8, "betas": (0.9, 0.99), "lr": 5e-5, "weight_decay": 5e-2}
        self.p_oversample_foreground = 0.6



if __name__ == "__main__":

    import lightning as L
    import os
    from batchgenerators.utilities.file_and_folder_operations import load_json
    from yucca.paths import get_preprocessed_data_path
    from yucca.pipeline.configuration.configure_task import TaskConfig
    from yucca.pipeline.configuration.configure_paths import get_path_config
    from yucca.pipeline.configuration.configure_callbacks import get_callback_config
    from yucca.pipeline.configuration.split_data import get_split_config
    from yucca.modules.data.augmentation.YuccaAugmentationComposer import YuccaAugmentationComposer
    from yucca.modules.data.augmentation.augmentation_presets import no_aug
    from yucca.modules.lightning_modules.YuccaLightningModule import YuccaLightningModule
    from yucca.modules.data.data_modules.YuccaDataModule import YuccaDataModule
    from yucca.documentation.templates.template_config import config

    config["plans"] = load_json(
        os.path.join(get_preprocessed_data_path(), config["task"], config["plans_name"], config["plans_name"] + "_plans.json")
    )

    task_config = TaskConfig(
        task=config.get("task"),
        continue_from_most_recent=True,
        experiment=config.get("experiment"),
        manager_name="",
        model_dimensions=config.get("dims"),
        model_name=config.get("model_name"),
        patch_based_training=True,
        planner_name=config.get("plans_name"),
        split_idx=config.get("split_idx"),
        split_method=config.get("split_method"),
        split_param=config.get("split_param"),
    )

    path_config = get_path_config(task_config=task_config, stage="fit")

    split_config = get_split_config(method=task_config.split_method, param=task_config.split_param, path_config=path_config)

    callback_config = get_callback_config(
        save_dir=path_config.save_dir,
        version_dir=path_config.version_dir,
        experiment=task_config.experiment,
        version=path_config.version,
        latest_ckpt_epochs=1,
        enable_logging=False,
        store_best_ckpt=False,
    )

    augmenter = YuccaAugmentationComposer(
        deep_supervision=config.get("deep_supervision"),
        patch_size=config.get("patch_size"),
        is_2D=True if config.get("dims") == "2D" else False,
        parameter_dict=no_aug,
        task_type_preset=config.get("task_type"),
    )

    model_module = YuccaLightningModule(
        config=config | task_config.lm_hparams() | path_config.lm_hparams() | callback_config.lm_hparams(),
        deep_supervision=config.get("deep_supervision"),
        optimizer_kwargs={"learning_rate": config.get("learning_rate"), "momentum": config.get("momentum")},
        loss_fn=config.get("loss_fn"),
    )

    data_module = YuccaDataModule(
        batch_size=config.get("batch_size"),
        patch_size=config.get("patch_size"),
        composed_train_transforms=augmenter.train_transforms,
        composed_val_transforms=augmenter.val_transforms,
        train_data_dir=path_config.train_data_dir,
        split_idx=task_config.split_idx,
        splits_config=split_config,
        task_type=config.get("task_type"),
    )

    trainer = L.Trainer(
        callbacks=callback_config.callbacks,
        default_root_dir=path_config.save_dir,
        limit_train_batches=2,
        limit_val_batches=2,
        logger=callback_config.loggers,
        precision="16",
        enable_progress_bar=True,
        max_epochs=config["max_epochs"],
        accelerator="cpu",
    )

    trainer.fit(
        model=model_module,
        datamodule=data_module,
        ckpt_path="last",
    )



def test_2D_networks_work():
    import torch
    from yucca.modules.networks.networks import MedNeXt, UNet, UNetR, UXNet, MultiResUNet
    from yucca.functional.utils.kwargs import filter_kwargs

    n_modalities = torch.randint(low=1, high=5, size=(1,))
    n_classes = torch.randint(low=1, high=5, size=(1,))
    patch_size = (32, 32)
    data = torch.randn((2, n_modalities, *patch_size))
    conv_op = torch.nn.Conv2d
    norm_op = torch.nn.InstanceNorm2d

    kwargs = {
        "input_channels": n_modalities,
        "num_classes": n_classes,
        "conv_op": conv_op,
        "norm_op": norm_op,
        "patch_size": patch_size,
    }
    networks = [MedNeXt, UNet, UNetR, UXNet, MultiResUNet]
    for network in networks:
        kw = filter_kwargs(network, kwargs)
        net = network(**kw)
        out = net(data)
        assert out is not None


def test_3D_networks_work():
    import torch
    from yucca.modules.networks.networks import MedNeXt, UNet, UNetR, UXNet, MultiResUNet
    from yucca.functional.utils.kwargs import filter_kwargs

    n_modalities = torch.randint(low=1, high=5, size=(1,))
    n_classes = torch.randint(low=1, high=5, size=(1,))
    patch_size = (32, 32, 32)
    data = torch.randn((2, n_modalities, *patch_size))
    conv_op = torch.nn.Conv3d
    norm_op = torch.nn.InstanceNorm3d

    kwargs = {
        "input_channels": n_modalities,
        "num_classes": n_classes,
        "conv_op": conv_op,
        "norm_op": norm_op,
        "patch_size": patch_size,
    }
    networks = [MedNeXt, UNet, UNetR, UXNet, MultiResUNet]
    for network in networks:
        kw = filter_kwargs(network, kwargs)
        net = network(**kw)
        out = net(data)
        assert out is not None



import lightning as L
import torch
import logging
import copy
from torchmetrics import MetricCollection
from torchmetrics.classification import Dice
from yucca.pipeline.preprocessing.YuccaPreprocessor import YuccaPreprocessor
from yucca.modules.optimization.loss_functions.deep_supervision import DeepSupervisionLoss
from yucca.modules.metrics.training_metrics import F1
from yucca.modules.optimization.loss_functions.nnUNet_losses import DiceCE
from yucca.functional.preprocessing import reverse_preprocessing


class BaseLightningModule(L.LightningModule):
    def __init__(
        self,
        model: torch.nn.Module,
        model_dimensions: str,
        num_classes: int,
        num_modalities: int,
        patch_size: tuple,
        plans: dict,
        deep_supervision: bool = False,
        disable_inference_preprocessing: bool = False,
        hparams_path: str = None,
        loss_fn: torch.nn.Module = DiceCE,
        loss_kwargs: dict = {
            "soft_dice_kwargs": {"apply_softmax": True},
        },
        lr_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.CosineAnnealingLR,
        lr_scheduler_kwargs: dict = {
            "T_max": 1000,
            "eta_min": 1e-9,
        },
        model_kwargs: dict = {},
        optimizer: torch.optim.Optimizer = torch.optim.SGD,
        optimizer_kwargs: dict = {
            "lr": 1e-3,
            "momentum": 0.9,
            "eps": 1e-4,
            "weight_decay": 3e-5,
        },
        preprocessor: YuccaPreprocessor = None,
        progress_bar: bool = False,
        sliding_window_overlap: float = 0.5,
        sliding_window_prediction: bool = True,
        step_logging: bool = False,
        test_time_augmentation: bool = False,
    ):
        super().__init__()
        # Extract parameters from the configurator
        self.num_classes = num_classes
        self.num_modalities = num_modalities
        self.hparams_path = hparams_path
        self.plans = plans
        self.model = model
        self.model_dimensions = model_dimensions
        self.model_kwargs = model_kwargs
        self.patch_size = patch_size
        self.preprocessor = preprocessor
        self.sliding_window_prediction = sliding_window_prediction

        # Loss, optimizer and scheduler parameters
        self.deep_supervision = deep_supervision
        self.disable_inference_preprocessing = disable_inference_preprocessing
        self.loss_fn = loss_fn
        self.loss_kwargs = loss_kwargs
        self.optim = optimizer
        self.optim_kwargs = optimizer_kwargs
        self.lr_scheduler = lr_scheduler
        self.lr_scheduler_kwargs = lr_scheduler_kwargs

        # Inference
        self.sliding_window_overlap = sliding_window_overlap
        self.test_time_augmentation = test_time_augmentation

        # Evaluation and logging
        if step_logging is True:  # Blame PyTorch lightning for this war crime
            self.step_logging = None
            self.epoch_logging = None
        else:
            self.step_logging = False
            self.epoch_logging = True

        self.progress_bar = progress_bar

        logging.info(f"{self.__class__.__name__} initialized")
        logging.info(f"Deep Supervision Enabled: {self.deep_supervision}")

        # If we are training we save params and then start training
        # Do not overwrite parameters during inference.
        self.save_hyperparameters(ignore=["model", "loss_fn", "lr_scheduler", "optimizer", "preprocessor"])

    def setup(self, stage):  # noqa: U100
        self.model = self.model(
            input_channels=self.num_modalities,
            num_classes=self.num_classes,
            **self.model_kwargs,
        )
        logging.info(f"Loading Model: {self.model.__class__.__name__} with kwargs: {self.model_kwargs}")

    def compute_metrics(self, metrics, output, target, ignore_index: int = 0):
        metrics = metrics(output, target)
        tmp = {}
        to_drop = []
        for key in metrics.keys():
            if metrics[key].numel() > 1:
                to_drop.append(key)
                for i, val in enumerate(metrics[key]):
                    if not i == ignore_index:
                        tmp[key + "_" + str(i)] = val
        for k in to_drop:
            metrics.pop(k)
        metrics.update(tmp)
        return metrics

    def configure_optimizers(self):
        self.loss_fn_train = self.loss_fn(**self.loss_kwargs)
        self.loss_fn_val = self.loss_fn(**self.loss_kwargs)

        # If deep_supervision is enabled we wrap our training loss (and potentially specify weights)
        # We leave the validation loss as is, as deep_supervision is not used for validation.
        if self.deep_supervision:
            self.loss_fn_train = DeepSupervisionLoss(self.loss_fn_train, weights=None)

        self.optim = self.optim(self.model.parameters(), **self.optim_kwargs)
        self.lr_scheduler = self.lr_scheduler(self.optim, **self.lr_scheduler_kwargs)
        return {"optimizer": self.optim, "lr_scheduler": self.lr_scheduler}

    def forward(self, inputs):
        return self.model(inputs)

    def load_state_dict(self, state_dict, *args, **kwargs):
        # First we filter out layers that have changed in size
        # This is often the case in the output layer.
        # If we are finetuning on a task with a different number of classes
        # than the pretraining task, the # output channels will have changed.
        old_params = copy.deepcopy(self.state_dict())
        state_dict = {
            k: v for k, v in state_dict.items() if (k in old_params) and (old_params[k].shape == state_dict[k].shape)
        }
        rejected_keys_new = [k for k in state_dict.keys() if k not in old_params]
        rejected_keys_shape = [k for k in state_dict.keys() if old_params[k].shape != state_dict[k].shape]
        rejected_keys_data = []

        # Here there's also potential to implement custom loading functions.
        # E.g. to load 2D pretrained models into 3D by repeating or something like that.

        # Now keep track of the # of layers with succesful weight transfers
        successful = 0
        unsuccessful = 0
        super().load_state_dict(state_dict, *args, **kwargs)
        new_params = self.state_dict()
        for param_name, p1, p2 in zip(old_params.keys(), old_params.values(), new_params.values()):
            # If more than one param in layer is NE (not equal) to the original weights we've successfully loaded new weights.
            if p1.data.ne(p2.data).sum() > 0:
                successful += 1
            else:
                unsuccessful += 1
                if param_name not in rejected_keys_new and param_name not in rejected_keys_shape:
                    rejected_keys_data.append(param_name)

        logging.warn(f"Succesfully transferred weights for {successful}/{successful+unsuccessful} layers")
        logging.warn(
            f"Rejected the following keys:\n"
            f"Not in old dict: {rejected_keys_new}.\n"
            f"Wrong shape: {rejected_keys_shape}.\n"
            f"Post check not succesful: {rejected_keys_data}."
        )

        return successful

    def on_fit_start(self):
        self.train_metrics = MetricCollection(
            {
                "train/dice": Dice(num_classes=self.num_classes, ignore_index=0 if self.num_classes > 1 else None),
                "train/F1": F1(num_classes=self.num_classes, ignore_index=0 if self.num_classes > 1 else None, average=None),
            },
        )
        self.val_metrics = MetricCollection(
            {
                "val/dice": Dice(num_classes=self.num_classes, ignore_index=0 if self.num_classes > 1 else None),
                "val/F1": F1(num_classes=self.num_classes, ignore_index=0 if self.num_classes > 1 else None, average=None),
            },
        )

    def on_predict_start(self):
        if self.disable_inference_preprocessing is False:
            self.preprocessor = self.preprocessor(self.hparams_path)

    def on_before_batch_transfer(self, batch, dataloader_idx):
        if self.trainer.predicting is True:
            if self.disable_inference_preprocessing is False:
                batch["data"], batch["data_properties"] = self.preprocessor.preprocess_case_for_inference(
                    images=batch["data_paths"],
                    patch_size=self.patch_size,
                    ext=batch["extension"],
                    sliding_window_prediction=self.sliding_window_prediction,
                )
        return super().on_before_batch_transfer(batch, dataloader_idx)

    def predict_step(self, batch, batch_idx, dataloader_idx=0):  # noqa: U100
        logits = self.model.predict(
            data=batch["data"],
            mode=self.model_dimensions,
            mirror=self.test_time_augmentation,
            overlap=self.sliding_window_overlap,
            patch_size=self.patch_size,
            sliding_window_prediction=self.sliding_window_prediction,
        )
        if self.disable_inference_preprocessing:
            logits, data_properties = reverse_preprocessing(
                crop_to_nonzero=self.plans["crop_to_nonzero"],
                images=logits,
                image_properties=batch["data_properties"],
                n_classes=self.num_classes,
                transpose_forward=self.plans["transpose_forward"],
                transpose_backward=self.plans["transpose_backward"],
            )
        else:
            logits, data_properties = self.preprocessor.reverse_preprocessing(
                logits, batch["data_properties"], num_classes=self.num_classes
            )
        return {"logits": logits, "properties": data_properties, "case_id": batch["case_id"]}

    def training_step(self, batch, batch_idx):  # noqa: U100
        inputs, target = batch["image"], batch["label"]
        output = self(inputs)
        loss = self.loss_fn_train(output, target)

        if self.deep_supervision:
            # If deep_supervision is enabled output and target will be a list of (downsampled) tensors.
            # We only need the original ground truth and its corresponding prediction which is always the first entry in each list.
            output = output[0]
            target = target[0]

        metrics = self.compute_metrics(self.train_metrics, output, target)
        self.log_dict(
            {"train/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )
        return loss

    def validation_step(self, batch, batch_idx):  # noqa: U100
        inputs, target = batch["image"], batch["label"]
        output = self(inputs)
        loss = self.loss_fn_val(output, target)

        metrics = self.compute_metrics(self.val_metrics, output, target)
        self.log_dict(
            {"val/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )



from batchgenerators.utilities.file_and_folder_operations import load_json
import argparse


def get_dice_and_nsd(path):
    file = load_json(path)
    dice_vals = []
    nsd_vals = []
    for i in file["mean"].keys():
        if i in [0, "0"]:
            continue
        dice_vals.append(file["mean"][i]["Dice"])
        nsd_vals.append(file["mean"][i]["Average Surface Distance"])

    print("\nDICE:", dice_vals)
    print(f"=AVERAGE{tuple(dice_vals)} \n")
    print("NSD: ", nsd_vals)
    print(f"=AVERAGE{tuple(nsd_vals)} \n")


def main():
    parser = argparse.ArgumentParser()

    # Required Arguments #
    parser.add_argument("--path")
    args = parser.parse_args()

    get_dice_and_nsd(args.path)


if __name__ == "__main__":
    main()



from torchvision import transforms
from yucca.functional.array_operations.matrix_ops import get_max_rotated_size
from yucca.modules.data.augmentation.transforms import (
    AddBatchDimension,
    AdditiveNoise,
    BiasField,
    Blur,
    CollectMetadata,
    ConvertLabelsToRegions,
    CopyImageToLabel,
    DownsampleSegForDS,
    Gamma,
    GibbsRinging,
    Masking,
    Mirror,
    MotionGhosting,
    MultiplicativeNoise,
    Normalize,
    RemoveBatchDimension,
    SimulateLowres,
    Spatial,
)


class YuccaAugmentationComposer:
    def __init__(
        self,
        patch_size: list | tuple,
        deep_supervision: bool = False,
        is_2D: bool = False,
        parameter_dict: dict = {},
        task_type_preset: str = "segmentation",
        labels: dict = None,
        regions: dict[str, dict] = None,
    ):
        self._pre_aug_patch_size = None
        self.deep_supervision = deep_supervision
        self.labels = labels
        self.regions = regions

        self.setup_default_params(is_2D, patch_size)
        self.apply_task_type_specific_preset(task_type_preset)
        self.overwrite_params(parameter_dict)
        self.train_transforms = self.compose_train_transforms()
        self.val_transforms = self.compose_val_transforms()

    def setup_default_params(self, is_2D, patch_size):
        """
        Default probabilities (all ON by default)
        For augmentations with AUG_P_PER_SAMPLE, AUG_P_PER_CHANNEL and AUG_P_PER_AXIS they are applied in the following order:
                for sample in dataset: # iterate over the batch
         if p_per_sample > np.random.uniform():
           for channel in sample: # iterate over the channel/modalities
             if p_per_channel > np.random.uniform():
               for axis in channel: # iterate over the h,w,d dimensions
                 if p_per_axis > np.random.uniform():
                   sample[channel, axis] = apply_aug(sample[channel, axis])
        Therefore, to enable any augmentation the AUG_P_PER_SAMPLE must be > 0.
        """
        print("Composing Transforms")

        # Define if the cropping performed during the spatial transform is random or always centered
        #   We have ALREADY selected a random crop when we prepared the case in the dataloader
        #   so unless you know what you're doing this should be disabled to avoid border artifacts
        self.random_crop = False
        self.mask_image_for_reconstruction = False
        self.patch_size = patch_size
        self.cval = "min"  # can be an int, float or a str in ['min', 'max']
        self.clip_to_input_range = False  # ensures no augmentations go beyond the input range of the image/patch
        self.normalize = False

        # label/segmentation transforms
        self.skip_label = False
        self.label_dtype = int
        self.copy_image_to_label = False
        self.convert_labels_to_regions = False

        # default augmentation probabilities
        self.additive_noise_p_per_sample = 0.2
        self.biasfield_p_per_sample = 0.33
        self.blurring_p_per_sample = 0.2
        self.blurring_p_per_channel = 0.5
        self.elastic_deform_p_per_sample = 0.33
        self.gamma_p_per_sample = 0.2
        self.gamma_p_invert_image = 0.05
        self.gibbs_ringing_p_per_sample = 0.2
        self.mirror_p_per_sample = 0.0
        self.mirror_p_per_axis = 0.33
        self.motion_ghosting_p_per_sample = 0.2
        self.multiplicative_noise_p_per_sample = 0.2
        self.rotation_p_per_sample = 0.2
        self.rotation_p_per_axis = 0.66
        self.scale_p_per_sample = 0.2
        self.simulate_lowres_p_per_sample = 0.2
        self.simulate_lowres_p_per_channel = 0.5
        self.simulate_lowres_p_per_axis = 0.33

        # default augmentation values
        self.additive_noise_mean = (0.0, 0.0)
        self.additive_noise_sigma = (1e-3, 1e-4)

        self.blurring_sigma = (0.0, 1.0)

        self.elastic_deform_alpha = (200, 600)
        self.elastic_deform_sigma = (20, 30)

        self.gamma_range = (0.5, 2.0)

        self.gibbs_ringing_cut_freq = (96, 129)
        self.gibbs_ringing_axes = (0, 2) if is_2D else (0, 3)

        self.mask_ratio = 0.5

        self.mirror_axes = (0, 1) if is_2D else (0, 1, 2)

        self.motion_ghosting_alpha = (0.85, 0.95)
        self.motion_ghosting_num_reps = (2, 11)
        self.motion_ghosting_axes = (0, 2) if is_2D else (0, 3)

        self.multiplicative_noise_mean = (0, 0)
        self.multiplicative_noise_sigma = (1e-3, 1e-4)

        self.normalization_scheme = "volume_wise_znorm"

        self.rotation_x = (-30.0, 30.0)
        self.rotation_y = (-0.0, 0.0) if is_2D else (-30.0, 30.0)
        self.rotation_z = (-0.0, 0.0) if is_2D else (-30.0, 30.0)

        self.scale_factor = (0.9, 1.1)

        self.simulate_lowres_zoom_range = (0.5, 1.0)

    @property
    def pre_aug_patch_size(self):
        # First check if any spatial transforms are included
        if self.elastic_deform_p_per_sample > 0 or self.rotation_p_per_sample > 0 or self.scale_p_per_sample > 0:
            self._pre_aug_patch_size = get_max_rotated_size(self.patch_size)
        return self._pre_aug_patch_size

    def apply_task_type_specific_preset(self, preset):
        if preset == "segmentation":
            if self.regions is not None:
                self.convert_labels_to_regions = True
        elif preset == "classification":
            self.skip_label = True
        elif preset == "self-supervised":
            self.skip_label = True
            self.copy_image_to_label = True
            self.mask_image_for_reconstruction = True
        else:
            raise ValueError(f"{preset} is not a valid `task_type_preset`.")

    def overwrite_params(self, parameter_dict):
        for key, value in parameter_dict.items():
            setattr(self, key, value)

    def compose_train_transforms(self):
        tr_transforms = transforms.Compose(
            [
                CollectMetadata(),
                AddBatchDimension(),
                # augmentations
                Spatial(
                    patch_size=self.patch_size,
                    crop=True,
                    random_crop=self.random_crop,
                    cval=self.cval,
                    clip_to_input_range=self.clip_to_input_range,
                    p_deform_per_sample=self.elastic_deform_p_per_sample,
                    deform_sigma=self.elastic_deform_sigma,
                    deform_alpha=self.elastic_deform_alpha,
                    p_rot_per_sample=self.rotation_p_per_sample,
                    p_rot_per_axis=self.rotation_p_per_axis,
                    x_rot_in_degrees=self.rotation_x,
                    y_rot_in_degrees=self.rotation_y,
                    z_rot_in_degrees=self.rotation_z,
                    p_scale_per_sample=self.scale_p_per_sample,
                    scale_factor=self.scale_factor,
                    skip_label=self.skip_label,
                ),
                AdditiveNoise(
                    p_per_sample=self.additive_noise_p_per_sample,
                    mean=self.additive_noise_mean,
                    sigma=self.additive_noise_sigma,
                    clip_to_input_range=self.clip_to_input_range,
                ),
                Blur(
                    p_per_sample=self.blurring_p_per_sample,
                    p_per_channel=self.blurring_p_per_channel,
                    sigma=self.blurring_sigma,
                    clip_to_input_range=self.clip_to_input_range,
                ),
                MultiplicativeNoise(
                    p_per_sample=self.multiplicative_noise_p_per_sample,
                    mean=self.multiplicative_noise_mean,
                    sigma=self.multiplicative_noise_sigma,
                    clip_to_input_range=self.clip_to_input_range,
                ),
                MotionGhosting(
                    p_per_sample=self.motion_ghosting_p_per_sample,
                    alpha=self.motion_ghosting_alpha,
                    num_reps=self.motion_ghosting_num_reps,
                    axes=self.motion_ghosting_axes,
                    clip_to_input_range=self.clip_to_input_range,
                ),
                GibbsRinging(
                    p_per_sample=self.gibbs_ringing_p_per_sample,
                    cut_freq=self.gibbs_ringing_cut_freq,
                    axes=self.gibbs_ringing_axes,
                    clip_to_input_range=self.clip_to_input_range,
                ),
                SimulateLowres(
                    p_per_sample=self.simulate_lowres_p_per_sample,
                    p_per_channel=self.simulate_lowres_p_per_channel,
                    p_per_axis=self.simulate_lowres_p_per_axis,
                    zoom_range=self.simulate_lowres_zoom_range,
                    clip_to_input_range=self.clip_to_input_range,
                ),
                BiasField(
                    p_per_sample=self.biasfield_p_per_sample,
                    clip_to_input_range=self.clip_to_input_range,
                ),
                Gamma(
                    p_per_sample=self.gamma_p_per_sample,
                    p_invert_image=self.gamma_p_invert_image,
                    gamma_range=self.gamma_range,
                    clip_to_input_range=self.clip_to_input_range,
                ),
                Mirror(
                    p_per_sample=self.mirror_p_per_sample,
                    axes=self.mirror_axes,
                    p_mirror_per_axis=self.mirror_p_per_axis,
                    skip_label=self.skip_label,
                ),
                Normalize(normalize=self.normalize, scheme=self.normalization_scheme),
                # seg
                ConvertLabelsToRegions(
                    convert_labels_to_regions=self.convert_labels_to_regions, labels=self.labels, regions=self.regions
                ),
                CopyImageToLabel(copy=self.copy_image_to_label),
                DownsampleSegForDS(deep_supervision=self.deep_supervision),
                # mae
                Masking(mask=self.mask_image_for_reconstruction, pixel_value=self.cval, ratio=self.mask_ratio),
                RemoveBatchDimension(),
            ]
        )
        return tr_transforms

    def compose_val_transforms(self):
        val_transforms = transforms.Compose(
            [
                AddBatchDimension(),
                ConvertLabelsToRegions(
                    convert_labels_to_regions=self.convert_labels_to_regions, regions=self.regions, labels=self.labels
                ),
                CopyImageToLabel(copy=self.copy_image_to_label),
                Masking(mask=self.mask_image_for_reconstruction, pixel_value=self.cval, ratio=self.mask_ratio),
                RemoveBatchDimension(),
            ]
        )
        return val_transforms

    def lm_hparams(self):
        hparams = {
            "augmentation_parameters": {
                "deep_supervision": self.deep_supervision,
                "pre_aug_patch_size": self.pre_aug_patch_size,
                "random_crop": self.random_crop,
                "cval": self.cval,
                "mask_image_for_reconstruction": self.mask_image_for_reconstruction,
                "patch_size": self.patch_size,
                "skip_label": self.skip_label,
                "label_dtype": self.label_dtype,
                "copy_image_to_label": self.copy_image_to_label,
                "convert_labels_to_regions": self.convert_labels_to_regions,
                "additive_noise_p_per_sample": self.additive_noise_p_per_sample,
                "additive_noise_mean": self.additive_noise_mean,
                "additive_noise_sigma": self.additive_noise_sigma,
                "biasfield_p_per_sample": self.biasfield_p_per_sample,
                "blurring_p_per_sample": self.blurring_p_per_sample,
                "blurring_sigma": self.blurring_sigma,
                "blurring_p_per_channel": self.blurring_p_per_channel,
                "elastic_deform_p_per_sample": self.elastic_deform_p_per_sample,
                "elastic_deform_alpha": self.elastic_deform_alpha,
                "elastic_deform_sigma": self.elastic_deform_sigma,
                "gamma_p_per_sample": self.gamma_p_per_sample,
                "gamma_p_invert_image": self.gamma_p_invert_image,
                "gamma_range": self.gamma_range,
                "gibbs_ringing_p_per_sample": self.gibbs_ringing_p_per_sample,
                "gibbs_ringing_cut_freq": self.gibbs_ringing_cut_freq,
                "gibbs_ringing_axes": self.gibbs_ringing_axes,
                "mask_ratio": self.mask_ratio,
                "mirror_p_per_sample": self.mirror_p_per_sample,
                "mirror_p_per_axis": self.mirror_p_per_axis,
                "mirror_axes": self.mirror_axes,
                "motion_ghosting_p_per_sample": self.motion_ghosting_p_per_sample,
                "motion_ghosting_alpha": self.motion_ghosting_alpha,
                "motion_ghosting_num_reps": self.motion_ghosting_num_reps,
                "motion_ghosting_axes": self.motion_ghosting_axes,
                "multiplicative_noise_p_per_sample": self.multiplicative_noise_p_per_sample,
                "multiplicative_noise_mean": self.multiplicative_noise_mean,
                "multiplicative_noise_sigma": self.multiplicative_noise_sigma,
                "rotation_p_per_sample": self.rotation_p_per_sample,
                "rotation_p_per_axis": self.rotation_p_per_axis,
                "rotation_x": self.rotation_x,
                "rotation_y": self.rotation_y,
                "rotation_z": self.rotation_z,
                "scale_p_per_sample": self.scale_p_per_sample,
                "scale_factor": self.scale_factor,
                "simulate_lowres_p_per_sample": self.simulate_lowres_p_per_sample,
                "simulate_lowres_p_per_channel": self.simulate_lowres_p_per_channel,
                "simulate_lowres_p_per_axis": self.simulate_lowres_p_per_axis,
                "simulate_lowres_zoom_range": self.simulate_lowres_zoom_range,
            }
        }
        return hparams


if __name__ == "__main__":
    from yucca.modules.data.augmentation.augmentation_presets import all_always
    import numpy as np

    print("NO CLIPPING")
    augs = all_always
    augs["clip_to_input_range"] = False
    x = YuccaAugmentationComposer(patch_size=(32, 32, 32), parameter_dict=augs)
    ttf = x.train_transforms
    mns = []
    for i in range(100):
        arr = np.random.randn(1, 32, 32, 32)
        arr = (arr - arr.min()) / (arr.max() - arr.min())
        mn = arr.min()
        mx = arr.max()
        arr = ttf({"image": arr})["image"]
        # print(f"BEFORE: min: {mn} max: {mx} AFTER: min: {arr.min()} max: {arr.max()}")
        mns.append(arr.min())
    print("no clip min:", np.min(mns))

    print("WITH CLIPPING")
    augs = all_always
    augs["clip_to_input_range"] = True
    x = YuccaAugmentationComposer(patch_size=(32, 32, 32), parameter_dict=augs)
    ttf = x.train_transforms
    mns = []
    for i in range(100):
        arr = np.random.randn(1, 32, 32, 32)
        arr = (arr - arr.min()) / (arr.max() - arr.min())
        mn = arr.min()
        mx = arr.max()
        arr = ttf({"image": arr})["image"]
        # print(f"BEFORE: min: {mn} max: {mx} AFTER: min: {arr.min()} max: {arr.max()}")
        mns.append(arr.min())
    print("with clip min:", np.min(mns))



if __name__ == "__main__":

    import re
    import os
    import numpy as np
    import torch
    from batchgenerators.utilities.file_and_folder_operations import (
        subfiles,
        join,
        save_pickle,
        maybe_mkdir_p as ensure_dir_exists,
        save_json,
    )
    from yucca.paths import get_raw_data_path, get_preprocessed_data_path
    from yucca.documentation.templates.template_config import config
    from yucca.functional.preprocessing import preprocess_case_for_training_with_label, preprocess_case_for_inference
    from yucca.functional.utils.loading import read_file_to_nifti_or_np
    from yucca.functional.planning import make_plans_file, add_stats_to_plans_post_preprocessing

    raw_images_dir = join(get_raw_data_path(), config["task"], "imagesTr")
    raw_labels_dir = join(get_raw_data_path(), config["task"], "labelsTr")
    test_raw_images_dir = join(get_raw_data_path(), config["task"], "imagesTs")

    target_dir = join(get_preprocessed_data_path(), config["task"], config["plans_name"])
    test_target_dir = join(get_preprocessed_data_path(), config["task"] + "_test", config["plans_name"])

    ensure_dir_exists(target_dir)
    ensure_dir_exists(test_target_dir)

    plans = make_plans_file(
        allow_missing_modalities=False,
        crop_to_nonzero=config["crop_to_nonzero"],
        norm_op=config["norm_op"],
        classes=[0, 1, 2],
        plans_name=config["plans_name"],
        modalities=config["modalities"],
        task_type=config["task_type"],
    )

    # Preprocess the training data
    subjects = [file[: -len(config["extension"])] for file in subfiles(raw_labels_dir, join=False) if not file.startswith(".")]

    for sub in subjects:
        images = [
            image_path
            for image_path in subfiles(raw_images_dir)
            if re.search(re.escape(sub) + "_" + r"\d{3}" + ".", os.path.split(image_path)[-1])
        ]
        images = [read_file_to_nifti_or_np(image) for image in images]
        label = read_file_to_nifti_or_np(join(raw_labels_dir, sub + config["extension"]))
        images, label, image_props = preprocess_case_for_training_with_label(
            images=images,
            label=label,
            normalization_operation=[plans["normalization_scheme"] for _ in config["modalities"]],
            allow_missing_modalities=False,
            enable_cc_analysis=False,
            crop_to_nonzero=plans["crop_to_nonzero"],
        )
        images = np.vstack((np.array(images), np.array(label)[np.newaxis]), dtype=np.float32)

        save_path = join(target_dir, sub)
        np.save(save_path + ".npy", images)
        save_pickle(image_props, save_path + ".pkl")

    plans = add_stats_to_plans_post_preprocessing(plans=plans, directory=target_dir)
    save_json(plans, join(target_dir, config["plans_name"] + "_plans.json"), sort_keys=False)

    # Preprocess the test data
    subjects = [
        file[: -len("_000" + config["extension"])]
        for file in subfiles(test_raw_images_dir, join=False)
        if not file.startswith(".")
    ]

    for sub in subjects:
        images = [
            image_path
            for image_path in subfiles(test_raw_images_dir)
            if re.search(re.escape(sub) + "_" + r"\d{3}" + ".", os.path.split(image_path)[-1])
        ]
        images = [read_file_to_nifti_or_np(image) for image in images]
        images, image_props = preprocess_case_for_inference(
            crop_to_nonzero=plans["crop_to_nonzero"],
            keep_aspect_ratio=plans["keep_aspect_ratio_when_using_target_size"],
            images=images,
            intensities=None,
            normalization_scheme=[plans["normalization_scheme"] for _ in config["modalities"]],
            patch_size=config["patch_size"],
            target_size=plans["target_size"],
            target_spacing=plans["target_spacing"],
            target_orientation=plans["target_coordinate_system"],
            transpose_forward=plans["transpose_forward"],
        )
        save_path = join(test_target_dir, sub)
        torch.save(images, save_path + ".pt")
        save_pickle(image_props, save_path + ".pkl")

    save_json(plans, join(test_target_dir, config["plans_name"] + "_plans.json"), sort_keys=False)



import numpy as np


def verify_labels_are_equal(expected_labels: np.ndarray, actual_labels: np.ndarray, id=""):
    if expected_labels.dtype == actual_labels.dtype:
        if np.all(np.isin(actual_labels, expected_labels)):
            return True
        else:
            print(f"Unexpected labels found for {id} \n" f"expected: {expected_labels} \n" f"found: {actual_labels}")
            return False
    else:
        print(
            "make sure reference and target is the same dtype before comparing the labels. \n"
            f"reference is: {expected_labels.dtype} and target is: {actual_labels.dtype}"
        )
        return False


def verify_array_shape_is_equal(reference: np.ndarray, target: np.ndarray, id=""):
    if np.all(reference.shape == target.shape):
        return True
    else:
        print(f"Sizes do not match for {id}" f"Image is: {reference.shape} while the label is {target.shape}")
        return False


def verify_shape_is_equal(reference, target, id=""):
    if np.all(reference == target):
        return True
    else:
        print(f"Sizes do not match for {id}" f"Image is: {reference.shape} while the label is {target.shape}")
        return False



from yucca.modules.data.augmentation.transforms.BiasField import BiasField
from yucca.modules.data.augmentation.transforms.Blur import Blur
from yucca.modules.data.augmentation.transforms.convert_labels_to_regions import ConvertLabelsToRegions
from yucca.modules.data.augmentation.transforms.copy_image_to_label import CopyImageToLabel
from yucca.modules.data.augmentation.transforms.cropping_and_padding import CropPad
from yucca.modules.data.augmentation.transforms.formatting import (
    RemoveBatchDimension,
    RemoveSegChannelAxis,
    NumpyToTorch,
    AddBatchDimension,
    CollectMetadata,
)
from yucca.modules.data.augmentation.transforms.Gamma import Gamma
from yucca.modules.data.augmentation.transforms.Ghosting import MotionGhosting
from yucca.modules.data.augmentation.transforms.Masking import Masking
from yucca.modules.data.augmentation.transforms.Mirror import Mirror
from yucca.modules.data.augmentation.transforms.Noise import AdditiveNoise, MultiplicativeNoise
from yucca.modules.data.augmentation.transforms.normalize import Normalize
from yucca.modules.data.augmentation.transforms.Ringing import GibbsRinging
from yucca.modules.data.augmentation.transforms.sampling import DownsampleSegForDS
from yucca.modules.data.augmentation.transforms.SimulateLowres import SimulateLowres
from yucca.modules.data.augmentation.transforms.Spatial import Spatial



from .MedNeXt import MedNeXt
from .MultiResUNet import MultiResUNet
from .TinyUNet import TinyUNet
from .UNet import UNet
from .UNetR import UNetR
from .UXNet import UXNet

networks = [MedNeXt, MultiResUNet, TinyUNet, UNet, UNetR, UXNet]
